@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
+ * Copyright (c) 2019-21, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
  *
  * NVIDIA CORPORATION and its licensors retain all intellectual property
  * and proprietary rights in and to this software, related documentation
@@ -17,10 +17,12 @@
 #pragma once
 
 #define CUTENSOR_MAJOR 1 //!< cuTensor major version.
-#define CUTENSOR_MINOR 2 //!< cuTensor minor version.
-#define CUTENSOR_PATCH 2 //!< cuTensor patch version.
+#define CUTENSOR_MINOR 4 //!< cuTensor minor version.
+#define CUTENSOR_PATCH 0 //!< cuTensor patch version.
+#define CUTENSOR_VERSION (CUTENSOR_MAJOR * 10000 + CUTENSOR_MINOR * 100 + CUTENSOR_PATCH)
 
 #include <stdint.h>
+#include <stdio.h>
 #include <cuda_runtime.h>
 
 #include <cutensor/types.h>
@@ -252,18 +254,32 @@ cutensorStatus_t cutensorInitTensorDescriptor(const cutensorHandle_t* handle,
  *
  * Supported data-type combinations are:
  *
- * \verbatim
- * |    typeA   |    typeB   |    typeC   | typeCompute |
- * | -----------|------------|------------|-------------|
- * | CUDA_R_16F | CUDA_R_16F | CUDA_R_16F | CUDA_R_16F  |
- * | CUDA_R_32F | CUDA_R_32F | CUDA_R_32F | CUDA_R_32F  |
- * | CUDA_R_64F | CUDA_R_64F | CUDA_R_64F | CUDA_R_64F  |
- * | CUDA_C_32F | CUDA_C_32F | CUDA_C_32F | CUDA_C_32F  |
- * | CUDA_C_64F | CUDA_C_64F | CUDA_C_64F | CUDA_C_64F  |
- * | CUDA_R_16F | CUDA_R_16F | CUDA_R_16F | CUDA_R_32F  |
- * | CUDA_R_32F | CUDA_R_32F | CUDA_R_16F | CUDA_R_32F  |
- * | CUDA_R_64F | CUDA_R_64F | CUDA_R_32F | CUDA_R_64F  |
- * | CUDA_C_64F | CUDA_C_64F | CUDA_C_32F | CUDA_C_64F  |
+ * \verbatim embed:rst:leading-asterisk
+ * +---------------+---------------+---------------+---------------+
+ * |     typeA     |     typeB     |     typeC     |  typeScalar   |
+ * +===============+===============+===============+===============+
+ * |  CUDA_R_16F   |  CUDA_R_16F   |  CUDA_R_16F   |  CUDA_R_16F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_R_16F   |  CUDA_R_16F   |  CUDA_R_16F   |  CUDA_R_32F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_R_16BF  |  CUDA_R_16BF  |  CUDA_R_16BF  |  CUDA_R_16BF  |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_R_16BF  |  CUDA_R_16BF  |  CUDA_R_16BF  |  CUDA_R_32F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_R_32F   |  CUDA_R_32F   |  CUDA_R_32F   |  CUDA_R_32F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_R_64F   |  CUDA_R_64F   |  CUDA_R_64F   |  CUDA_R_64F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_C_32F   |  CUDA_C_32F   |  CUDA_C_32F   |  CUDA_C_32F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_C_64F   |  CUDA_C_64F   |  CUDA_C_64F   |  CUDA_C_64F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_R_32F   |  CUDA_R_32F   |  CUDA_R_16F   |  CUDA_R_32F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_R_64F   |  CUDA_R_64F   |  CUDA_R_32F   |  CUDA_R_64F   |
+ * +---------------+---------------+---------------+---------------+
+ * |  CUDA_C_64F   |  CUDA_C_64F   |  CUDA_C_32F   |  CUDA_C_64F   |
+ * +---------------+---------------+---------------+---------------+
  * \endverbatim
  *
  * \param[in] handle Opaque handle holding cuTENSOR's library context.
@@ -284,11 +300,12 @@ cutensorStatus_t cutensorInitTensorDescriptor(const cutensorHandle_t* handle,
  * \param[in] modeD Array (in host memory) of size descD->numModes that holds the names of the modes of D. The modeD[i] corresponds to extent[i] and stride[i] of the cutensorInitTensorDescriptor.
  * \param[in] opAB Element-wise binary operator (see \f$\Phi_{AB}\f$ above).
  * \param[in] opABC Element-wise binary operator (see \f$\Phi_{ABC}\f$ above).
- * \param[in] typeScalar Scalar type for the intermediate computation.
+ * \param[in] typeScalar Denotes the data type for the scalars alpha, beta, and gamma. Moreover, typeScalar determines the data type that is used throughout the computation.
  * \param[in] stream The cuda stream.
  * \retval CUTENSOR_STATUS_SUCCESS The operation completed successfully.
  * \retval CUTENSOR_STATUS_NOT_INITIALIZED if the handle is not initialized.
  * \retval CUTENSOR_STATUS_INVALID_VALUE if some input data is invalid (this typically indicates an user error).
+ * \retval CUTENSOR_STATUS_ARCH_MISMATCH if the device is either not ready, or the target architecture is not supported.
  * \remarks calls asynchronous functions, no reentrant, and thread-safe
  *
  */
@@ -324,7 +341,7 @@ cutensorStatus_t cutensorElementwiseTrinary(const cutensorHandle_t* handle,
  * \param[in] stream The cuda stream.
  * \retval CUTENSOR_STATUS_NOT_SUPPORTED if the combination of data types or operations is not supported
  * \retval CUTENSOR_STATUS_INVALID_VALUE if tensor dimensions or modes have an illegal value
- * \retval CUTENSOR_STATUS_SUCCESS The operation completed successfully.without error
+ * \retval CUTENSOR_STATUS_SUCCESS The operation completed successfully without error
  * \retval CUTENSOR_STATUS_NOT_INITIALIZED if the handle is not initialized.
  * \remarks calls asynchronous functions, no reentrant, and thread-safe
  */
@@ -365,7 +382,7 @@ cutensorStatus_t cutensorElementwiseBinary(const cutensorHandle_t* handle,
  * \param[in] stream The CUDA stream.
  * \retval CUTENSOR_STATUS_NOT_SUPPORTED if the combination of data types or operations is not supported
  * \retval CUTENSOR_STATUS_INVALID_VALUE if tensor dimensions or modes have an illegal value
- * \retval CUTENSOR_STATUS_SUCCESS The operation completed successfully.without error
+ * \retval CUTENSOR_STATUS_SUCCESS The operation completed successfully without error
  * \retval CUTENSOR_STATUS_NOT_INITIALIZED if the handle is not initialized.
  * \remarks calls asynchronous functions, no reentrant, and thread-safe
  */
@@ -401,7 +418,7 @@ cutensorStatus_t cutensorPermutation(const cutensorHandle_t* handle,
  * \param[in] typeCompute Datatype of for the intermediate computation of typeCompute T = A * B.
  * \retval CUTENSOR_STATUS_NOT_SUPPORTED if the combination of data types or operations is not supported
  * \retval CUTENSOR_STATUS_INVALID_VALUE if tensor dimensions or modes have an illegal value
- * \retval CUTENSOR_STATUS_SUCCESS The operation completed successfully.without error
+ * \retval CUTENSOR_STATUS_SUCCESS The operation completed successfully without error
  * \retval CUTENSOR_STATUS_NOT_INITIALIZED if the handle is not initialized.
  */
 cutensorStatus_t cutensorInitContractionDescriptor(const cutensorHandle_t* handle,
@@ -508,6 +525,7 @@ cutensorStatus_t cutensorContractionGetWorkspace(const cutensorHandle_t* handle,
  * \retval CUTENSOR_STATUS_SUCCESS If a viable candidate has been found.
  * \retval CUTENSOR_STATUS_NOT_SUPPORTED If no viable candidate could be found.
  * \retval CUTENSOR_STATUS_NOT_INITIALIZED if the handle is not initialized.
+ * \retval CUTENSOR_STATUS_INSUFFICIENT_WORKSPACE if The provided workspace was insufficient.
  * \retval CUTENSOR_STATUS_INVALID_VALUE if some input data is invalid (this typically indicates an user error).
  */
 cutensorStatus_t cutensorInitContractionPlan(const cutensorHandle_t* handle,
@@ -525,19 +543,38 @@ cutensorStatus_t cutensorInitContractionPlan(const cutensorHandle_t* handle,
  *
  * Supported data-type combinations are:
  *
- * \verbatim
- * |    typeA   |    typeB   |    typeC   | typeCompute |
- * | -----------|------------|------------|-------------|
- * | CUDA_R_16F | CUDA_R_16F | CUDA_R_16F | CUDA_R_32F  |
- * | CUDA_R_32F | CUDA_R_32F | CUDA_R_32F | CUDA_R_32F  |
- * | CUDA_R_32F | CUDA_R_32F | CUDA_R_32F | CUDA_R_16F  |
- * | CUDA_R_64F | CUDA_R_64F | CUDA_R_64F | CUDA_R_64F  |
- * | CUDA_R_64F | CUDA_R_64F | CUDA_R_64F | CUDA_R_32F  |
- * | CUDA_C_32F | CUDA_C_32F | CUDA_C_32F | CUDA_C_32F  |
- * | CUDA_C_64F | CUDA_C_64F | CUDA_C_64F | CUDA_C_64F  |
- * | CUDA_C_64F | CUDA_C_64F | CUDA_C_64F | CUDA_C_32F  |
- * | CUDA_R_64F | CUDA_C_64F | CUDA_C_64F | CUDA_C_64F  |
- * | CUDA_C_64F | CUDA_R_64F | CUDA_C_64F | CUDA_C_64F  |
+ * \verbatim embed:rst:leading-asterisk
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |     typeA     |     typeB     |     typeC     |        typeCompute      | Tensor Core |
+ * +===============+===============+===============+=========================+=============+
+ * |  CUDA_R_16F   |  CUDA_R_16F   |  CUDA_R_16F   |  CUTENSOR_COMPUTE_32F   | Volta+      |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_R_16BF  |  CUDA_R_16BF  |  CUDA_R_16BF  |  CUTENSOR_COMPUTE_32F   | Ampere+     |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_R_32F   |  CUDA_R_32F   |  CUDA_R_32F   |  CUTENSOR_COMPUTE_32F   | No          |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_R_32F   |  CUDA_R_32F   |  CUDA_R_32F   |  CUTENSOR_COMPUTE_TF32  | Ampere+     |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_R_32F   |  CUDA_R_32F   |  CUDA_R_32F   |  CUTENSOR_COMPUTE_16BF  | Ampere+     |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_R_32F   |  CUDA_R_32F   |  CUDA_R_32F   |  CUTENSOR_COMPUTE_16F   | Volta+      |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_R_64F   |  CUDA_R_64F   |  CUDA_R_64F   |  CUTENSOR_COMPUTE_64F   | Ampere+     |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_R_64F   |  CUDA_R_64F   |  CUDA_R_64F   |  CUTENSOR_COMPUTE_32F   | No          |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_C_32F   |  CUDA_C_32F   |  CUDA_C_32F   |  CUTENSOR_COMPUTE_32F   | No          |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_C_32F   |  CUDA_C_32F   |  CUDA_C_32F   |  CUTENSOR_COMPUTE_TF32  | Ampere+     |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_C_64F   |  CUDA_C_64F   |  CUDA_C_64F   |  CUTENSOR_COMPUTE_64F   | Ampere+     |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_C_64F   |  CUDA_C_64F   |  CUDA_C_64F   |  CUTENSOR_COMPUTE_32F   | No          |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_R_64F   |  CUDA_C_64F   |  CUDA_C_64F   |  CUTENSOR_COMPUTE_64F   | No          |
+ * +---------------+---------------+---------------+-------------------------+-------------+
+ * |  CUDA_C_64F   |  CUDA_R_64F   |  CUDA_C_64F   |  CUTENSOR_COMPUTE_64F   | No          |
+ * +---------------+---------------+---------------+-------------------------+-------------+
  * \endverbatim
  *
  * \param[in] handle Opaque handle holding cuTENSOR's library context.
@@ -559,6 +596,9 @@ cutensorStatus_t cutensorInitContractionPlan(const cutensorHandle_t* handle,
  * \retval CUTENSOR_STATUS_INVALID_VALUE if some input data is invalid (this typically indicates an user error).
  * \retval CUTENSOR_STATUS_SUCCESS The operation completed successfully.
  * \retval CUTENSOR_STATUS_NOT_INITIALIZED if the handle is not initialized.
+ * \retval CUTENSOR_STATUS_ARCH_MISMATCH if the plan was created for a different device than the currently active device.
+ * \retval CUTENSOR_STATUS_INSUFFICIENT_DRIVER if the driver is insufficient.
+ * \retval CUTENSOR_STATUS_CUDA_ERROR if some unknown CUDA error has occurred (e.g., out of memory).
  */
 cutensorStatus_t cutensorContraction(const cutensorHandle_t* handle, 
                                      const cutensorContractionPlan_t* plan,
@@ -584,20 +624,32 @@ cutensorStatus_t cutensorContractionMaxAlgos(int32_t* maxNumAlgos);
  *
  * This function is also able to perform partial reductions; for instance: C[i,j] = alpha * A[k,j,i]; in this case only elements along the k-mode are contracted.
  *
- * It the binary opReduce operator provides extra control over what kind of a reduction
+ * The binary opReduce operator provides extra control over what kind of a reduction
  * ought to be perfromed. For instance, opReduce == CUTENSOR_OP_ADD reduces element of A
  * via a summation while CUTENSOR_OP_MAX would find the largest element in A.
  *
  * Supported data-type combinations are:
  *
- * \verbatim
- * |    typeA   |    typeB   |    typeC   | typeCompute |
- * | -----------|------------|------------|-------------|
- * | CUDA_R_16F | CUDA_R_16F | CUDA_R_16F | CUDA_R_32F  |
- * | CUDA_R_32F | CUDA_R_32F | CUDA_R_32F | CUDA_R_32F  |
- * | CUDA_R_64F | CUDA_R_64F | CUDA_R_64F | CUDA_R_64F  |
- * | CUDA_C_32F | CUDA_C_32F | CUDA_C_32F | CUDA_C_32F  |
- * | CUDA_C_64F | CUDA_C_64F | CUDA_C_64F | CUDA_C_64F  |
+ * \verbatim embed:rst:leading-asterisk
+ * +---------------+---------------+---------------+-------------------------+
+ * |     typeA     |     typeB     |     typeC     |       typeCompute       |
+ * +===============+===============+===============+=========================+
+ * | `CUDA_R_16F`  | `CUDA_R_16F`  | `CUDA_R_16F`  | `CUTENSOR_COMPUTE_16F`  |
+ * +---------------+---------------+---------------+-------------------------+
+ * | `CUDA_R_16F`  | `CUDA_R_16F`  | `CUDA_R_16F`  | `CUTENSOR_COMPUTE_32F`  |
+ * +---------------+---------------+---------------+-------------------------+
+ * | `CUDA_R_16BF` | `CUDA_R_16BF` | `CUDA_R_16BF` | `CUTENSOR_COMPUTE_16BF` |
+ * +---------------+---------------+---------------+-------------------------+
+ * | `CUDA_R_16BF` | `CUDA_R_16BF` | `CUDA_R_16BF` | `CUTENSOR_COMPUTE_32F`  |
+ * +---------------+---------------+---------------+-------------------------+
+ * | `CUDA_R_32F`  | `CUDA_R_32F`  | `CUDA_R_32F`  | `CUTENSOR_COMPUTE_32F`  |
+ * +---------------+---------------+---------------+-------------------------+
+ * | `CUDA_R_64F`  | `CUDA_R_64F`  | `CUDA_R_64F`  | `CUTENSOR_COMPUTE_64F`  |
+ * +---------------+---------------+---------------+-------------------------+
+ * | `CUDA_C_32F`  | `CUDA_C_32F`  | `CUDA_C_32F`  | `CUTENSOR_COMPUTE_32F`  |
+ * +---------------+---------------+---------------+-------------------------+
+ * | `CUDA_C_64F`  | `CUDA_C_64F`  | `CUDA_C_64F`  | `CUTENSOR_COMPUTE_64F`  |
+ * +---------------+---------------+---------------+-------------------------+
  * \endverbatim
  *
  * \param[in] handle Opaque handle holding cuTENSOR's library context.
@@ -693,6 +745,54 @@ size_t cutensorGetVersion();
  */
 size_t cutensorGetCudartVersion();
 
+/**
+ * \brief This function sets the logging callback routine.
+ * \param[in] callback Pointer to a callback function. Check cutensorLoggerCallback_t.
+ */
+cutensorStatus_t cutensorLoggerSetCallback(cutensorLoggerCallback_t callback);
+
+/**
+ * \brief This function sets the logging output file.
+ * \param[in] file An open file with write permission.
+ */
+cutensorStatus_t cutensorLoggerSetFile(FILE* file);
+
+/**
+ * \brief This function opens a logging output file in the given path.
+ * \param[in] logFile Path to the logging output file.
+ */
+cutensorStatus_t cutensorLoggerOpenFile(const char* logFile);
+
+/**
+ * \brief This function sets the value of the logging level.
+ * \param[in] level Log level, should be one of the following:
+ *                  0.  Off
+ *                  1.  Errors
+ *                  2.  Performance Trace
+ *                  3.  Performance Hints
+ *                  4.  Heuristics Trace
+ *                  5.  API Trace
+ */
+cutensorStatus_t cutensorLoggerSetLevel(int32_t level);
+
+/**
+ * \brief This function sets the value of the log mask.
+ * \param[in] mask Log mask, the bitwise OR of the following:
+ *                 0.  Off
+ *                 1.  Errors
+ *                 2.  Performance Trace
+ *                 4.  Performance Hints
+ *                 8.  Heuristics Trace
+ *                 16. API Trace
+ *
+ */
+cutensorStatus_t cutensorLoggerSetMask(int32_t mask);
+
+/**
+ * \brief This function disables logging for the entire run.
+ */
+cutensorStatus_t cutensorLoggerForceDisable();
+
 #if defined(__cplusplus)
 }
 #endif /* __cplusplus */
