@@ -2,14 +2,13 @@
 set -ex
 
 test -f $PREFIX/include/cutensor.h
+test -f $PREFIX/include/cutensorMg.h
 test -f $PREFIX/include/cutensor/types.h
 test -f $PREFIX/lib/libcutensor.so
-TEST_LINKER_FLAGS=""
-if [[ $target_platform == linux-ppc64le ]]; then
-    TEST_LINKER_FLAGS+=" -L/usr/local/cuda/lib64 -lcudart"
-fi
-${GCC} test_load_elf.c -std=c99 -Werror -ldl $TEST_LINKER_FLAGS -o test_load_elf
+test -f $PREFIX/lib/libcutensorMg.so
+${GCC} test_load_elf.c -std=c99 -Werror -ldl -o test_load_elf
 ./test_load_elf $PREFIX/lib/libcutensor.so
+./test_load_elf $PREFIX/lib/libcutensorMg.so
 
 NVCC_FLAGS=""
 # Workaround __ieee128 error; see https://github.com/LLNL/blt/issues/341
